/****************************************************************************
*****************************************************************************
**   The Array Extender                                                    **
*****************************************************************************
*****************************************************************************
**                                                                         **
**  By: Alan Kreutzer  --                                                  **
**  May 2002                                                               **
**                                                                         **
**  Requires WinBatch version 2001 or later                                **
**                                                                         **
**  May be freely distrubuted.                                             **
**  This program is distributed with no warrenty, use at your own risk.    **
**                                                                         **
*****************************************************************************
****************************************************************************/
//#define EXTENDERVERSION           // Read Version from RC file,  #1001

#define INTERFACELEVEL  2           // extender interface level 
#define EXTENDER_FEATURE_LEVEL 1001 // extender feature level 

#define STRICT


/*---------------------------------------------------------------------------+
!                                   Includes                                 !
+---------------------------------------------------------------------------*/
#include <windows.h>
                           /* We need to include "stdio.h" from the       */
                           /* C library so we can use "sprintf" to        */
                           /* convert floating point numbers to strings.  */
#include <stdio.h>
                           /* Special includes for WinBatch extenders     */
#include "wwwpform.h"
#include "addons.h"




/*---------------------------------------------------------------------------+
!                                                                            !
!       Definitions used when manipulating arrays.                           !
!                                                                            !
+---------------------------------------------------------------------------*/
#define VARTYPE_UNDEF      0           // These are bitmapped, var can be both
#define VARTYPE_INT        1           // Numeric and string, must check bits
#define VARTYPE_STRING     2
#define VARTYPE_FILE       5           // FILE and INT bits set
#define VARTYPE_UNINIT     8
#define VARTYPE_OLEOBJECT 17           // OLE AND INT bits set
#define VARTYPE_FLOATNUM  32           // floating point
#define VARTYPE_BINARY    65           // BINARY and INT bits set
#define VARTYPE_XOLEOBJECT 129         /// an OLE object that has been temporarily camouflaged
#define VARTYPE_ARRAY     256          /// an array

typedef struct arrayvariable
{
  double Aflt;                         // 8-byte floating point number
  LONG   Ax;                           // if int variable, value is here
  int    Ainittype;                    // 0=uninitialized 1=int 2=string 3=both  yes can have both sometimes
  DWORD  Astrptr;                      // pointer to string space for string
} arrvar;


typedef struct resultants2
{
  union u
  {
    struct s                           // for normal variables
    {
      double flt;                      // 8-byte floating point number
      LONG   x;                        // if int variable, value is here
    };
    struct ss                          // for VARTYPE_ARRAY
    {
      WORD wDim[5];                    // array dimension sizes
      BYTE cDims;                      // number of array dimensions
      BYTE cFlags;                     // array flags
    };
  };
  int   inittype;                      // 0=uninitialized 1=int 2=string 3=both  yes can have both sometimes
  union uu
  {
    DWORD strptr;                      // pointer to string space for string
    LPSTR lpBuffer;                    // pointer to locked array data buffer for VARTYPE_ARRAY
  };
  WORD  vorigin;                       // index to var area is derived from var, or FFFF
} resultvar2;

typedef arrvar FAR *LPARRVAR;

typedef resultvar2 FAR * LPRESULTVAR2;
   
/*****************************************************************************
*                                                                            *
*       Common data block.                                                   *
*                                                                            *
*       This structure is passed to all called routines to                   *
*       pass all common data between them.                                   *
*                                                                            *
*       To return data to WinBatch,  set the "Type" variable to              *
*       A (Array) ... S (String) ... L (Long)  ... or E (Error).             *
*       Then put the data to be returned in the appropriate place.           *
*                                                                            *
*       A ... Create the array in "Array"                                    *
*       L ... Put the number in "Long"                                       *
*       S ... Put pointer to string in "String"                              *
*       E ... Put the error number in "Long"  (Error number points to        *
*             an error message in the "RC" file.                             *'
*             Put -1 (Normal error) or -3 (Fatal error) in "ErrorType"       *
*                                                                            *
*       "lpfnDllServices" and "hBatData" are values passed as parameters     *
*       by the wil interpereter.  The DllServices callback is used for       *
*       some array functions.  (It's probably useful for lots of other       *
*       things, if only we knew.)  We copy the values to the Common block    *
*       so that called routines call use them, too.                          *
*                                                                            *
*       The "ToFree" array contains pointers to any memory buffers           *
*       we borrow.  All borrowed memory will be freed before exit.           *
*                                                                            *
*****************************************************************************/
const long ToFreeMAX = 8;

typedef struct xCommonStruct
{
     long        Type;       /* A = Array, S = String, L = Long, E = Error. */
     long        Long;       /* Numeric value (when return = 'L'            */
     double      Float;      /* Floating point(when return = 'F'            */
     char        *String;    /* String        (when return = 'S' or 'E')    */
     long        ErrorType;  /* Error type... -1 (Normal) or -3 (Fatal)     */
     resultvar2  Array;      /* Array         (when return = 'A')           */

     DLLSERVICES lpfnDllServices;    /* Helper routine from WinBatch        */
     HGLOBAL     hBatData;           /* Passed from WinBatch                */

     long        ToFreeTop;
     void       *ToFree[ToFreeMAX];

} xCommon;

/*****************************************************************************
*                                                                            *
*               Function Prototypes                                          *
*                                                                            *
*****************************************************************************/
void  ArrayDuplicate(xCommon *Common, LPRESULTVAR2 ArrIn);
void  ArrayizeBytes (xCommon *Common, char *TheString);
void  ArrayResize   (xCommon *Common, long nArgCount, LPVIPERVAR lpvvArg);
void  ArraySorter   (xCommon *Common, LPRESULTVAR2 ArrIn, BOOL Descending, BOOL NoCase);
char *ArrayToList   (xCommon *Common, LPRESULTVAR2 ArrIn, char Delim);

char *AppendToString(char *A, char *B);
char *GetBuffer(xCommon *Common, long Size);

long  CompareAscending   (LPARRVAR A, LPARRVAR B);
long  CompareDescending  (LPARRVAR A, LPARRVAR B);
long  CompareAscendingNC (LPARRVAR A, LPARRVAR B);
long  CompareDescendingNC(LPARRVAR A, LPARRVAR B);

long  CompareStrings  (char *A, char *B, BOOL NoCase);
long  CompareLong2String(long L, char *S);
long  CompareString2Long(char *S, long L);

long  AtoL(char *A);

LPSTR GetArrayString(xCommon *Common, LPARRVAR ArrayItem); 
DWORD PutArrayString(xCommon *Common, LPARRVAR lpArray, char *TheString);
long  ArrayItemType(LPARRVAR ArrayItem);
LPARRVAR ArrayElement(LPRESULTVAR2 arr, WORD x0, WORD x1, WORD x2, WORD x3, WORD x4);

/*---------------------------------------------------------------------------+
!                Debug Functions                                             !
+---------------------------------------------------------------------------*/
void FAR PASCAL MBSS(char *Title, char *Text);  
void FAR PASCAL MBSN(char *Title, long Number);
void FAR PASCAL MBSF(char *Title, double Flt);
void FAR PASCAL MBSX(char *Title, long Number);



/*---------------------------------------------------------------------------+
!                           Global Variables                                 !
+---------------------------------------------------------------------------*/
         /*--------- Superstitious data field to bypass/ignore bugs in
         !    certain versions of Windows that cause the beginning 
         !    of the data area to be overwritten.  This field should 
         !    be the first data area defined and linked. 
         +------------------*/
char szBugFiller[] = "ABCDEFGHIJKLMNOPQRSTUVWXYZ";

HINSTANCE   hinstDLL;                    // DLL Instance saved



        /*------------------------------------------------------+
        !       Case conversion table.                          !
        !       (Converts upper to lower case)                  !
        +------------------------------------------------------*/
#define LOWERCASE(C) LowerTable[(unsigned char) C]
unsigned char LowerTable[] =
    {0x00,0x01,0x02,0x03,0x04,0x05,0x06,0x07,0x08,0x09,0x0a,0x0b,0x0c,0x0d,0x0e,0x0f,
     0x10,0x11,0x12,0x13,0x14,0x15,0x16,0x17,0x18,0x19,0x1a,0x1b,0x1c,0x1d,0x1e,0x1f,
     0x20,0x21,0x22,0x23,0x24,0x25,0x26,0x27,0x28,0x29,0x2a,0x2b,0x2c,0x2d,0x2e,0x2f,
     0x30,0x31,0x32,0x33,0x34,0x35,0x36,0x37,0x38,0x39,0x3a,0x3b,0x3c,0x3d,0x3e,0x3f,
     0x40,0x61,0x62,0x63,0x64,0x65,0x66,0x67,0x68,0x69,0x6a,0x6b,0x6c,0x6d,0x6e,0x6f,
     0x70,0x71,0x72,0x73,0x74,0x75,0x76,0x77,0x78,0x79,0x7a,0x5b,0x5c,0x5d,0x5e,0x5f,
     0x60,0x61,0x62,0x63,0x64,0x65,0x66,0x67,0x68,0x69,0x6a,0x6b,0x6c,0x6d,0x6e,0x6f,
     0x70,0x71,0x72,0x73,0x74,0x75,0x76,0x77,0x78,0x79,0x7a,0x7b,0x7c,0x7d,0x7e,0x7f,
     0x80,0x81,0x82,0x83,0x84,0x85,0x86,0x87,0x88,0x89,0x9a,0x8b,0x9c,0x8d,0x8e,0x8f, 
     0x90,0x91,0x92,0x93,0x94,0x95,0x96,0x97,0x98,0x99,0x9a,0x9b,0x9c,0x9d,0x9e,0xFf, 
     0xA0,0xA1,0xA2,0xA3,0xA4,0xA5,0xA6,0xA7,0xA8,0xA9,0xAa,0xAb,0xAc,0xAd,0xAe,0xAf,
     0xB0,0xB1,0xB2,0xB3,0xB4,0xB5,0xB6,0xB7,0xB8,0xB9,0xBa,0xBb,0xBc,0xBd,0xBe,0xBf, 
     0xE0,0xE1,0xE2,0xE3,0xE4,0xE5,0xE6,0xE7,0xE8,0xE9,0xEa,0xEb,0xEc,0xEd,0xEe,0xEf,
     0xF0,0xF1,0xF2,0xF3,0xF4,0xF5,0xF6,0xD7,0xF8,0xF9,0xFa,0xFb,0xFc,0xFd,0xFe,0xDf,
     0xE0,0xE1,0xE2,0xE3,0xE4,0xE5,0xE6,0xE7,0xE8,0xE9,0xEa,0xEb,0xEc,0xEd,0xEe,0xEf,
     0xF0,0xF1,0xF2,0xF3,0xF4,0xF5,0xF6,0xF7,0xF8,0xF9,0xFa,0xFb,0xFc,0xFd,0xFe,0xFf}; 





/***************************************************************************
*                                                                          *
*                                                                          *
*                  Command Extension Table                                 *
*                                                                          *
*                                                                          *
***************************************************************************/
tablestruct CommandExtTable[] =
{
//----------------------------------------------------------------------
// ID   Type       Mask              Params/Value, Size, Name
//----------------------------------------------------------------------
  {10, FUNCTION, 0x0000000000000000, 0x00000001L, 10, "ArrayExtenderVersion" }, 
  {11, FUNCTION, 0x0000000000000003, 0x00000101L,  9, "ArraySort"            },
  {12, FUNCTION, 0x0000000000000003, 0x00000101L, 11, "ArraySortNC"          },
  {13, FUNCTION, 0x0000000000000003, 0x00000001L, 14, "ArrayDuplicate"       },
  {14, FUNCTION, 0x00000000000000F3, 0x00000403L, 11, "ArrayResize"          }, 
  {15, FUNCTION, 0x0000000000000001, 0x00000001L, 13, "ArrayizeBytes"        },
  {16, FUNCTION, 0x0000000000000013, 0x00000101L, 11, "ArrayToList"          },
  
//--------------------------------------------------------------------------
  {-1,        0,                  0, 0x00000000,   0, ""                }
};




/*******************************************************************************
********************************************************************************
**                                                                            **
**                                                                            **
**                               DllMain                                      **
**                                                                            **
**                                                                            **
********************************************************************************
*******************************************************************************/
BOOL WINAPI DllMain(HANDLE hInst, DWORD dwReason, LPVOID lpReserved)
{

  hinstDLL = hInst;
  UNREFERENCED_PARAMETER(dwReason);
  UNREFERENCED_PARAMETER(lpReserved);
  return (TRUE);
}




/*******************************************************************************
********************************************************************************
**                                                                            **
**                                                                            **
**                        WILExtenderQuery                                    **
**                                                                            **
**           Tells the WilDll about our interface level.                      **
**                                                                            **
**                                                                            **
********************************************************************************
*******************************************************************************/
DllExport LONG FAR PASCAL WILExtenderQuery
 (short  nRequest,                     // type of information being requested
  LPVOID lpvInfo,                      // information buffer
  LONG   nCount)                       // additional data
{
  char szWork[256];
  LONG        i;

  switch (nRequest)
  {
/*----------------------------------------------------------------------+
!   case 1 ... identify yourself to a caller.                           !
+----------------------------------------------------------------------*/
    case 1:                      
    {
      if ((lpvInfo) && (nCount > 0))
      {
        LoadString(hinstDLL, 1000, szWork, sizeof(szWork));
        lstrcpyn((LPSTR) lpvInfo, szWork, nCount);
      }
      LoadString(hinstDLL, 1001, szWork, sizeof(szWork));
      return AtoL(szWork);
    }

/*----------------------------------------------------------------------+
!   case 2 ... Tell WinBatch we use interface level 2                   !   
+----------------------------------------------------------------------*/
    case 2:                            // preferred interface version
    {
      for (i = 0; i < nCount; i++)
        if (((LPLONG) lpvInfo)[i] == INTERFACELEVEL)  // found one we like
          return (INTERFACELEVEL);

      return (-1);                     // didn't find any we like
    }


/*----------------------------------------------------------------------+
!    unknown requests                                           !  
+----------------------------------------------------------------------*/
    default:
      return (0);
  }
}




/*******************************************************************************
********************************************************************************
**                                                                            **
**                                                                            **
**                                                                            **
**                          WILExtender2                                      **
**                                                                            **
**                                                                            **
**                                                                            **
**   The main routine performs all extender functions, and various            **
**   initialization and cleanup tasks.                                        **
**                                                                            **
**                                                                            **
**                                                                            **
********************************************************************************
*******************************************************************************/

DllExport LONG FAR PASCAL WILExtender2
  (HWND           hWnd,                   // hWnd of parent application
  HINSTANCE       hInst,                  // hInstance of parent application
  LONG            ViperInterfaceVersion,  // WIL interface version
  short           CmdId,                  // identifies extender function being called
  LPVIPERCALLBACK lpViperCallback,        // pointer to return value callback function
  LPVIPERSTRUCT   lpViperStruct,          // pointer to information structure
  LPLONG          lpLong,                 // pointer to your 32 bits
  LPVIPERVAR      lpvvArg,                // pointer to parameter array
  LONG            nArgCount,              // parameter count for lpvvArg
  LPVIPERVAR      lpViperReturn)          // pointer to return value structure
{  
  long          J;
  long          K;
  xCommon       Common;
  
  UNREFERENCED_PARAMETER(hWnd);
  UNREFERENCED_PARAMETER(hInst);
  UNREFERENCED_PARAMETER(lpLong);
  
  /*------------------------------------------------+
  !    Initialize the Common block                  !
  +------------------------------------------------*/
  Common.lpfnDllServices = lpViperStruct->lpfnDllServices;
  Common.hBatData = lpViperStruct->hBatData;
  
  Common.ToFreeTop = 0;
  Common.ErrorType = -1;
  Common.Array.lpBuffer = NULL;  
    

  
     /***************************************************
     *                                                  *
     *    The "Big Switch."                             *    
     *                                                  *
     *    Calls from Winbatch include an integer value  *
     *    (CmdId) which selects a function from this    *
     *    switch.   Calls < 0 are WinBatch overhead.    *
     *    Calls > 0 are our functions.                  *
     *                                                  *
     ***************************************************/             
  switch (CmdId)         
  {
  
/*----------------------------------------------------------------------+
!  case -1 ... One time initialization routine.                         !
!                                                                       !
!              Return the number of functions  & constants defined.     !
+----------------------------------------------------------------------*/
    case -1:                         
    {      
       return ((sizeof(CommandExtTable) / sizeof(tablestruct)) - 1); 
    }

/*----------------------------------------------------------------------+
!  case -2 ... Called repeatedly as extender is being loaded.           !
!                                                                       !
!              On each call, return a function table entry.             !
+----------------------------------------------------------------------*/
    case -2:                          
    {
      LPTABLESTRUCT lpTable;
      char         *IName;
      char         *OName;

      lpTable = (LPTABLESTRUCT) lpvvArg;
      J = (UINT) nArgCount;

      if (CommandExtTable[J].ident == -1)
        return (0);

      if (lpTable)
      {
        lpTable->ident  = CommandExtTable[J].ident;
        lpTable->type   = CommandExtTable[J].type;
        lpTable->param1 = CommandExtTable[J].param1;
        lpTable->ulMask = CommandExtTable[J].ulMask;
        K = 0;
        IName = (char *) CommandExtTable[J].delim;
        OName = (char *) lpTable->delim;
        while ((*OName++ = LOWERCASE(*IName++)) != '\0') K++;
        lpTable->size = K;
      }
      return (1);
    }


/*----------------------------------------------------------------------+
!  case -3 .. Called once at shutdown. Can be used for cleanup.         !
+----------------------------------------------------------------------*/
    case -3:
    {
        return (1);
    }


/*----------------------------------------------------------------------+
!  case 10 ... ArrxVersion                                              !
!                                                                       !
!             Return extender version information.                      !
!                                                                       !
!              Parameter 0 = Version                                    !
!                        1 = Total Functions                            !
!                        2 = Total Constants                            !
+----------------------------------------------------------------------*/
    case 10:
    {       
         UINT uType;      

     
         if (lpvvArg[0].x == 0)             // Request 0 = Extender version
         {
             char VersionString[24];
             LoadString(hinstDLL, 1001, VersionString, 24);            
             Common.Type = 'L';
             Common.Long = AtoL(VersionString);
             break;
         } 
   
         if (lpvvArg[0].x == 1)            // total functions
             uType = FUNCTION;
         else 
            if (lpvvArg[0].x == 2)         // total constants
                uType = CONSTANT;
            else
            {
                Common.Type = 'E';          // Invalid parameter
                Common.Long = 101;
                break;
            }

        Common.Type = 'L';
        Common.Long = 0;
        for (K = 0; CommandExtTable[K].ident != -1; K++)
           if ((UINT) CommandExtTable[K].type == uType)
                Common.Long++;
        break;
    }

/*----------------------------------------------------------------------+
!  case 11 ... ArraySort(Array [,Direction])                            !
!                                                                       !
!         Sort an array on column 1.    (Case Sensitive)                !
!         Parameter 2 = @ASCENDING or @DESCENDING.                      !
!         (If Parameter 2 is omitted, the default is @ASCENDING)        !
+----------------------------------------------------------------------*/ 
    case 11:
    {      
        BOOL Descending = FALSE;
        
        Common.Type = 'L';
        Common.Long = 1;    
        
        if (nArgCount > 1) Descending = (BOOL) lpvvArg[1].x;
        
        ArraySorter(&Common, (LPRESULTVAR2) &lpvvArg[0], Descending, FALSE);
        break;
    }

/*----------------------------------------------------------------------+
!  case 12 ... ArraySortNC(Array)                                       !
!                                                                       !
!         Sort an array on column 1.    (Not case sensitive)            !
+----------------------------------------------------------------------*/ 
    case 12:
    {      
        BOOL Descending = 0;
        
        Common.Type = 'L';
        Common.Long = 1;    
        
        if (nArgCount > 1) Descending = (BOOL) lpvvArg[1].x;
                
        ArraySorter(&Common, (LPRESULTVAR2) &lpvvArg[0], Descending, TRUE);
        break;
    }
    
/*----------------------------------------------------------------------+
!  case 13 ... ArrayDuplicate(Array)                                    !
!                                                                       !
!       Returns a duplicate of the array.                               !
!                                                                       !
!       In WinBatch, B = ARRAY   doesn't create a duplicte of an array. !
!       It creates a pointer, so that A and B both point at the same    !
!       array, and any changes to one also changes the other.           !
!                                                                       !
!       ArrayDuplicate returns a true duplicate so you can change the   !
!       new array without changing the old one.                         !
+----------------------------------------------------------------------*/ 
    case 13:
    {    
        ArrayDuplicate(&Common, (LPRESULTVAR2) &lpvvArg[0]); 
        break;
    }    
    
/*----------------------------------------------------------------------+
!  case 14 ... ArrayResize(Array,InitialValue,1,[2,3,4,5])              !
!                                                                       !
!         Returns an array of the specified dimensions.                 !
!         The new array is initialized with the contents of the         !
!         passed array.  If the new array is smaller than the old       !
!         array, elements will be lost.  If it is larger, the new       !
!         elements are initialized with the passed "InitialValue"       !
!                                                                       !
!         The passed array is not modified.                             !
+----------------------------------------------------------------------*/ 
    case 14:
    {      
       ArrayResize(&Common,nArgCount,lpvvArg);
       break;                   
    }
   
    
/*----------------------------------------------------------------------+
!  case 15 ... ArrayizeBytes(String)                                    !
!                                                                       !
!       Returns an array, each member containing one byte from          !
!       the string.                                                     !
!                                                                       !
+----------------------------------------------------------------------*/
    case 15:
    {      
       ArrayizeBytes(&Common, lpvvArg[0].lpstr);         
       break;
    } 
    
    
/*----------------------------------------------------------------------+
!  case 16 ... ArrayToList(Array [, Delimiter])                         !
!                                                                       !
!       Converts an array to a delimited list.                          !
!                                                                       !
!       The first character of "Delimiter" is used as the list          !
!       delimiter.  If omitted, the default delimiter is @TAB.          !
!       You may specify "" as the delimiter to concatenate the          !
!       items in the array with no delimiter between them.              !                                                                !
!                                                                       !
!                                                                       !
+----------------------------------------------------------------------*/
    case 16:
    {     
         char *TheList;
         char Delim;
         
         if (nArgCount > 1)
             Delim = lpvvArg[1].lpstr[0];
         else
             Delim = '\t';

         TheList = ArrayToList(&Common, (LPRESULTVAR2) &lpvvArg[0], Delim);
         if (TheList != NULL)
         {
             Common.Type = 'S';
             Common.String = TheList;
         }
         break;
    } 
    
               
/*----------------------------------------------------------------------+
!   Default Case.  (We should never get here)                           !
+-----------------------------------------------------------------------*/
    default:      
    {
        Common.Type = 'E';
        Common.Long = 100;
        break;
    }
  }                   
  
  
            /*************************************
            *                                    *
            *       End of "Big Switch"          *
            *                                    *
            **************************************
 
 
 
 
 
/*-----------------------------------------------------------------------+
!+----------------------------------------------------------------------+!
!!                                                                      !!
!!   Return the value,  free allocated memory, and return.              !!
!!                                                                      !!
!+----------------------------------------------------------------------+!
+-----------------------------------------------------------------------*/ 
 switch (Common.Type)         
  { 
      case 'A':   // Return an array
    {
       lpViperReturn->inittype = VARTYPE_ARRAY;
       lpViperReturn->flt      = Common.Array.flt;
       lpViperReturn->x        = Common.Array.x;
       lpViperReturn->lpstr    = Common.Array.lpBuffer;
       J = lpViperCallback(RESULT_SUCCESS, ViperInterfaceVersion, lpViperStruct, lpViperReturn);
       break;
    }
  
    case 'L':  // Return a long integer
    {
       lpViperReturn->inittype=VARTYPE_INT;
       lpViperReturn->x=Common.Long;
       J = lpViperCallback(RESULT_SUCCESS, ViperInterfaceVersion, lpViperStruct, lpViperReturn);
       break;
    }   
    
    case 'S':   // Return a string
    {
       lpViperReturn->inittype=VARTYPE_STRING;
       lpViperReturn->lpstr=Common.String;
       J = lpViperCallback(RESULT_SUCCESS, ViperInterfaceVersion, lpViperStruct, lpViperReturn);    
       break;
    }

    case 'F':   // Return a floating point number
    {
       lpViperReturn->inittype=VARTYPE_FLOATNUM;
       lpViperReturn->flt=Common.Float;
       J = lpViperCallback(RESULT_SUCCESS, ViperInterfaceVersion, lpViperStruct, lpViperReturn);
       break;
    }   
        
    case 'E':  // Return an error message
    {
       char ErrorMessage[128];
       LoadString(hinstDLL, Common.Long, ErrorMessage,128);
       lpViperReturn->inittype=VARTYPE_STRING; 
       lpViperReturn->lpstr = ErrorMessage;
       J = lpViperCallback(Common.ErrorType, ViperInterfaceVersion, lpViperStruct, lpViperReturn);
       break;
    }
    
    default:  // Invalid return type.  (This better not happen!)
    {
       char ErrorMessage[128];
       LoadString(hinstDLL, 900, ErrorMessage, 128);
       lpViperReturn->inittype=VARTYPE_STRING; 
       lpViperReturn->lpstr = ErrorMessage;
       J = lpViperCallback(-3, ViperInterfaceVersion, lpViperStruct, lpViperReturn);
       break;
    }  
  }

  /*--------------------------------------+
  !        Free borrowed memory           !
  +--------------------------------------*/
  for (J=0; J<Common.ToFreeTop; J++) GlobalFree(Common.ToFree[J]);
  return J;
}

  /************************************
   *                                  *
   *   End of program WILExtender2    *
   *                                  *
   ***********************************/



/************************************************************************
*                                                                       *
*  *******************************************************************  *
*  *                                                                 *  *
*  *                                                                 *  *
*  *          Helper routines.                                       *  *
*  *                                                                 *  *
*  *                                                                 *  *
*  *******************************************************************  *
*                                                                       *
*************************************************************************/  
       
/***************************************************************************
*                                                                          *
*    Duplicate an array                                                    *
*                                                                          *
***************************************************************************/
void ArrayDuplicate(xCommon *Common, LPRESULTVAR2  ArrIn)
{
    LPRESULTVAR2  ArrOut; 
    LPARRVAR      From;
    LPARRVAR      To;
    char         *Str;
    long          J;
    long          ItemCount;

    Common->Type = 'A';
    ArrOut =  &Common->Array;
    ItemCount =  ArrIn->wDim[0];    
    for (J=1; J<ArrIn->cDims;J++) ItemCount = ItemCount * ArrIn->wDim[J];
    
    ArrOut->cDims      = ArrIn->cDims;
    ArrOut->wDim[0]    = ArrIn->wDim[0];
    ArrOut->wDim[1]    = ArrIn->wDim[1];
    ArrOut->wDim[2]    = ArrIn->wDim[2];
    ArrOut->wDim[3]    = ArrIn->wDim[3];
    ArrOut->wDim[4]    = ArrIn->wDim[4];
    ArrOut->cFlags     = ArrIn->cFlags;
    ArrOut->lpBuffer   = GetBuffer(Common, ItemCount * sizeof(arrvar));
    if (ArrOut->lpBuffer == NULL) return;

    From = (LPARRVAR) ArrIn->lpBuffer;
    To   = (LPARRVAR) ArrOut->lpBuffer;
    CopyMemory(To, From,ItemCount * sizeof(arrvar)); 
    
    for (J=0;J<ItemCount;J++)
    {
        if (To->Ainittype & VARTYPE_STRING)
        {
            Str = GetArrayString(Common, From);
            if (Str != NULL) 
            { 
                if (PutArrayString(Common, To, Str) == 0) break;
            }
        }
        From  += 1;
        To    += 1;
    }
    return;
}

/***************************************************************************
*                                                                          *
*    Convert a string to an array, with one character per item.            *
*                                                                          *
***************************************************************************/
void ArrayizeBytes(xCommon *Common, char *TheString)
{
       LPARRVAR      Next; 
       long          J;
       long          ByteCount;
       char         *Buffer;  
       char          Work[2];


       if (*TheString == '\0') TheString = " ";
       
       /*---------------------------------------------------------------+
       !  Create a buffer and make a copy of the string.                !
       !                                                                !
       !  It's not clear why this is necessary, but before this was     !
       !  done the routine was crashing when passed very long strings.  !
       +---------------------------------------------------------------*/                                                    
       ByteCount = lstrlen(TheString);
       Buffer = GetBuffer(Common,ByteCount+1);
       if (Buffer == NULL) return;
       lstrcpy(Buffer,TheString);      

       /*---------------------------------------------------------------+
       !   Get a buffer to hold the new array.                          !
       +---------------------------------------------------------------*/
       Common->Array.lpBuffer  = GetBuffer(Common, ByteCount * sizeof(arrvar));
       if (Common->Array.lpBuffer == NULL) return;
             
       /*---------------------------------------------------------------+
       !   Initialize the array.                                        !
       +---------------------------------------------------------------*/ 
       Common->Array.cDims      = 1;
       Common->Array.wDim[0]    = ByteCount;
       Common->Array.wDim[1]    = 0;
       Common->Array.wDim[2]    = 0;
       Common->Array.wDim[3]    = 0;
       Common->Array.wDim[4]    = 0;
       Common->Array.cFlags     = 0;
         
       /*---------------------------------------------------------------+
       !  Walk through the string character by character, inserting     !
       !  the characters into the array.                                !
       +---------------------------------------------------------------*/
       Next = (LPARRVAR) Common->Array.lpBuffer;
       Work[1] = '\0';
       for (J=0; J<ByteCount; J++)
       {
           Work[0] = Buffer[J];
           if (PutArrayString(Common, Next, Work) == 0) break;
           Next++; 
       }
       
       Common->Type = 'A';
       return;
}

/***************************************************************************
*                                                                          *
*    Array Resize.                                                         *
*                                                                          *
*    lpvvArg contains the parameters passed from WinBatch.                 *
*     0   = The Array                                                      *
*     1   = Fill Value (to initialize entries if the array gets bigger)    *
*     3-7 = Dimensions of the new array.                                   *
*                                                                          *
***************************************************************************/
void  ArrayResize(xCommon *Common, long nArgCount, LPVIPERVAR lpvvArg)
{  
    LPRESULTVAR2  ArrIn; 
    LPRESULTVAR2  ArrOut;  
    LPARRVAR      From;
    LPARRVAR      To;
    long          J;
    long          K;
    long          NewItemCount;
    long          MaxDim;
    BOOLEAN       Done;
    long          x[5];
    char         *Str;
      
    int           InitialType;
    LONG          InitialValue;
       
    Common->Type = 'A';
    ArrIn = (LPRESULTVAR2) &lpvvArg[0]; 
    ArrOut = &Common->Array;
    
    //----------------------------------------------------------------
    //    Get the dimensions of the new array.
    //----------------------------------------------------------------   
    J = 0;
    K = 2;
    while(J<5)
    {
        if (K >= nArgCount) ArrOut->wDim[J] = 0;
        else                ArrOut->wDim[J] = lpvvArg[K].x;
        J++;
        K++;
    }

    ArrOut->cDims = 0;
    for (J=0; J<5; J++)
    {
       if (ArrOut->wDim[J] == 0) break;
       ArrOut->cDims++;
    }
                 
    //----------------------------------------------------------------
    //  Compute the number of items in the new array.
    //---------------------------------------------------------------- 
    NewItemCount = ArrOut->wDim[0];    
    for (K=1; K < ArrOut->cDims; K++) 
       NewItemCount = NewItemCount * ArrOut->wDim[K];

    if (NewItemCount < 1)
    {
        Common->Type = 'E';
        Common->Long = 102;
        return;
    }
       
    //----------------------------------------------------------------
    // Get memory for the new array.
    //---------------------------------------------------------------- 
    ArrOut->lpBuffer  = GetBuffer(Common, NewItemCount * sizeof(arrvar));
    if (ArrOut->lpBuffer == NULL) return;

    //----------------------------------------------------------------
    // initialize everything in the new array with an invalid value
    // so we can tell, later, which ones we didn't copy from the
    // old array.
    //---------------------------------------------------------------- 
    To = (LPARRVAR) ArrOut->lpBuffer;
    for (K=0; K < NewItemCount; K++)
    {
        To->Ainittype = -1;
        To++;
    }

    //---------------------------------------------------------------- 
    //  Copy corresponding items from the old array to the new array.
    //---------------------------------------------------------------- 
    MaxDim = ((ArrIn->cDims<Common->Array.cDims)?ArrIn->cDims:Common->Array.cDims)-1;
    x[0] = x[1] = x[2] = x[3] = x[4] = 0;
    Done = FALSE;
    while(!Done)
    {
        From = ArrayElement(ArrIn,x[0],x[1],x[2],x[3],x[4]);
        if (From != NULL)
        {
           To = ArrayElement(ArrOut,x[0],x[1],x[2],x[3],x[4]);
           if (To != NULL)
           {
               CopyMemory(To,From,sizeof(arrvar)); 
               if (To->Ainittype & VARTYPE_STRING)
               {
                 Str = GetArrayString(Common, From);
                 if (Str != NULL) 
                 {
                    if (PutArrayString(Common, To, Str) == 0) break;
                 }
               }
           }
        }
        if (Common->Type == 'E') break;
        J=MaxDim;
        while (1)
        {
           x[J]++;
           if (x[J] < ArrOut->wDim[J]) break;
           if (J == 0){Done=TRUE; break;}
           x[J] = 0;
           J--;
        }
    } 

    //---------------------------------------------------------------- 
    //  Find all the items we didn't copy and set them to the  
    //  user's "initial value"
    //----------------------------------------------------------------  
    InitialType = lpvvArg[1].inittype;
    
    To = (LPARRVAR) ArrOut->lpBuffer;
    if (InitialType & VARTYPE_STRING)      
    {
        for (J=0;J<NewItemCount;J++)
        {
            if (To->Ainittype == -1) 
            {
               if (PutArrayString(Common, To, lpvvArg[1].lpstr) == 0) break;
            }
            To++;
        }
        return;
    }
    
    if (InitialType & VARTYPE_FLOATNUM)
    {
        for (K=0; K<NewItemCount; K++)
        {
            if (To->Ainittype == -1)
            {
                To->Ainittype = VARTYPE_FLOATNUM;
                To->Aflt = lpvvArg[1].flt;
            }
            To++;
        }
        return;
    }
    
    InitialValue = lpvvArg[1].x;
    if (!(InitialType & VARTYPE_INT)) 
    {
       InitialType = VARTYPE_INT;
       InitialValue = 0;
    }
    for (K=0; K<NewItemCount; K++)
    {
        if (To->Ainittype == -1)
        {
            To->Ax = InitialValue;
            To->Ainittype = InitialType;
        }
        To++;
    }      
    return;
}
/***************************************************************************
*                                                                          *
*    Array Sort Routine                                                    *
*                                                                          *
*    Parameter 2 is the array.                                             *
*    Parameter 3 = TRUE for a descending sort, FALSE for ascending.        *
*    Parameter 4 = TRUE for to ignore casem, FALSE for case sensitive.     *
*                                                                          *
***************************************************************************/
void ArraySorter(xCommon *Common, LPRESULTVAR2 ArrIn, BOOL Descending, BOOL NoCase)
{
    LPARRVAR First;
    LPARRVAR Low;
    LPARRVAR Next;

    BOOL     MustSwap[3];
    long     LowType;
    long     NextType;

    char    *LowString;
    char    *NextString;

    long     Compare;
    char    *Hold;
        
    long     ItemSize;
    long     ItemCount;
    long     ItemCountMinus1;
    
    long     X;
    long     Y;
    
    DOUBLE   D;
        
    MustSwap[0] = FALSE;    
    if (Descending)
    {                           // For a descending sort
        MustSwap[1] = TRUE;     // we swap items when the "low" item
        MustSwap[2] = FALSE;    // is less than the current item.
    }
    else
    {                           // For ascending sorts,
        MustSwap[1] = FALSE;    // we swap when the "low" item 
        MustSwap[2] = TRUE;     // is is greater than the current item.
    }

    //------------------------------------------------------------------
    // Compute the distance (in bytes) between the first two items
    //------------------------------------------------------------------
    First  = (LPARRVAR) ArrIn->lpBuffer;
    if ((Next = ArrayElement(ArrIn,1,0,0,0,0)) == NULL) return;
    ItemSize = (Next - First) * sizeof(arrvar);
        
    //-------------------------------------------------------------------
    // Number of items
    //-------------------------------------------------------------------
    ItemCount = ArrIn->wDim[0];
    ItemCountMinus1 = ItemCount-1;
        
    //-------------------------------------------------------------------
    // Get a buffer for temporary storage. big enough to hold one item
    //-------------------------------------------------------------------   
    Hold =  GetBuffer(Common,ItemSize);         // Borrow memory.
    if (Hold == NULL) return;   
        
    //----------------------------------------------------
    // Sort the array.
    //----------------------------------------------------
    for (X=0; X<ItemCountMinus1; X++)
    {
       First = (LPARRVAR) (ArrIn->lpBuffer + (ItemSize * X));
       
       Low = First;
       LowType = ArrayItemType(Low);
       if (LowType == VARTYPE_STRING) 
       {  
          LowString = GetArrayString(Common, Low);
          if (LowString == NULL) return;
       }
       
       Next = Low;           
       for (Y=X+1; Y<ItemCount; Y++)
       {
            Next = (LPARRVAR) (((char *) Next) + ItemSize);
            NextType = ArrayItemType(Next);
            if (NextType == VARTYPE_STRING) 
            { 
                NextString = GetArrayString(Common, Next);
                if (NextString == NULL) return;
            }
            
            Compare = 0;
            switch (LowType)
            {
                case -1:
                {
                     if (NextType != -1) Compare = 1;
                     break;
                }   
                  
                case VARTYPE_INT:
                {
                    if (NextType == VARTYPE_INT)
                    {
                        if (Low->Ax < Next->Ax)  {Compare = 1;break;}
                        if (Low->Ax > Next->Ax)  {Compare = 2;break;}
                        break;
                    }
                    if (NextType == VARTYPE_STRING)
                    {
                        Compare = CompareLong2String(Low->Ax, NextString);
                        break;
                    }
                    if (NextType == VARTYPE_FLOATNUM)
                    {
                        D = (double) Low->Ax;
                        if (D < Next->Aflt) Compare = 1;
                        if (D > Next->Aflt) Compare = 2;
                        break;
                    }
                    break;
                }
    
                case VARTYPE_FLOATNUM:
                {
                    if (NextType == VARTYPE_FLOATNUM)
                    {
                        if (Low->Aflt < Next->Aflt) Compare = 1;
                        if (Low->Aflt > Next->Aflt) Compare = 2;
                        break;
                    }
                    if (NextType == VARTYPE_INT)
                    {
                        D = (double) Next->Ax;
                        if (Low->Aflt < D) Compare = 1;
                        if (Low->Aflt > D) Compare = 2;
                        break;
                    }
                    break;
                } 

                case VARTYPE_STRING:
                {
                    if (NextType == VARTYPE_STRING)
                    {
                        Compare = CompareStrings(LowString, NextString, NoCase);
                        break;
                    }
                    if (NextType == VARTYPE_INT)
                    {
                        Compare = CompareString2Long(LowString, Next->Ax);
                        break;
                    } 
                    break;
                }      
            }
            if (MustSwap[Compare])
            {
                 Low = Next;
                 LowType = NextType;
                 LowString = NextString;
            }
       }                  
       
       if (Low != First)
       {
            CopyMemory(Hold,Low,ItemSize);
            CopyMemory(Low,First,ItemSize);
            CopyMemory(First,Hold,ItemSize);
       }
    }
    return;
}


/***************************************************************************
*                                                                          *
*    Array to List.                                                        *
*                                                                          *
*    Convert an array to a delimited list.  Returns a pointer to the list. *
*                                                                          *
*    DELIM parameter is the list delimiter.  (It may be \0 to concatenate  *
*    the items with no delimiter.)                                         *
*                                                                          *
***************************************************************************/
char *ArrayToList(xCommon *Common, LPRESULTVAR2 ArrIn, char Delim)
{
    LPARRVAR First;
    LPARRVAR Next;

    long     J;
    long     ListLength; 
    long     ItemCount;
    long     ItemOffset;
    long     ItemType;
    
    char    *Str;
    char    *Buffer;
    char    *Append;
    char     DelimString[2];
    char     Work[32];

    DelimString[0] = Delim;
    DelimString[1] = '\0';
    
    //------------------------------------------------------------------
    // Compute the distance (in bytes) between the first items
    //------------------------------------------------------------------
    First  = (LPARRVAR) ArrIn->lpBuffer;
    if ((Next = ArrayElement(ArrIn,1,0,0,0,0)) == NULL)
        ItemOffset = 0;
    else
        ItemOffset = (Next - First) * sizeof(arrvar);
        
    //-------------------------------------------------------------------
    // Number of items
    //-------------------------------------------------------------------
    ItemCount = ArrIn->wDim[0];
        
    //-------------------------------------------------------------------
    //  Scan the array and extimate the size of string we'll need.  
    //-------------------------------------------------------------------
    Next = First;
    ListLength = 2;
    for (J=0; J<ItemCount; J++)
    {
        ItemType = ArrayItemType(Next);
        if (ItemType == VARTYPE_STRING)
        {
           Str = GetArrayString(Common, Next);
           ListLength = ListLength + lstrlen(Str) +1;
        }
        else
        {
           ListLength = ListLength + 14;
        }
        Next = (LPARRVAR) (((char *) Next) + ItemOffset); 
    } 
    
    //-------------------------------------------------------------------
    //    Get a buffer where we can create the list.
    //-------------------------------------------------------------------   
    Buffer =  GetBuffer(Common,ListLength);         // Borrow memory.
    if (Buffer == NULL) return NULL;   
        
    //----------------------------------------------------
    //   Create the list.
    //----------------------------------------------------
    Next = First;
    Append = Buffer;
    *Append = '\0';
    
    for (J=0; J<ItemCount; J++)
    {
        ItemType = ArrayItemType(Next);
        if (ItemType == VARTYPE_STRING)
        {
            Str = GetArrayString(Common, Next);
        }
        else
        {
           if (ItemType == VARTYPE_FLOATNUM)
               sprintf(Work,"%g",Next->Aflt);
           else
               sprintf(Work,"%d",Next->Ax);
           Str = Work;
        }
        Append = AppendToString(Append,Str);
        Append = AppendToString(Append,DelimString);
        Next = (LPARRVAR) (((char *) Next) + ItemOffset); 
    } 

    if (Delim != '\0')          // No trailing delmiter
      if (Append > Buffer)
          *(--Append) = '\0';
              
    return Buffer;
}

/**************************************************************************
*                                                                         *
*    Append String B to String A.  (Parameter A should point at the end   *
*    of string A.) Return the address of the (new) end of string A.       *
*                                                                         *
**************************************************************************/
char *AppendToString(char *A, char *B)
{
    while (1)
    {
        *A = *B;
        if (*A == '\0') return A;
        A++;
        B++;
    }
}
/****************************************************************************
*                                                                           *
*     Get a memory buffer.                                                  *
*                                                                           *
*     Return value is the buffer address, or NULL if there was an error.    *
*                                                                           *
*     (If there was an error, we have set the flags in the Common block     *
*      to return an error message.)                                         *
*                                                                           *
*     Buffer addresses are saved in the Common->ToFree array, so they       *
*     will be freed before we exit.                                         *
*                                                                           *
****************************************************************************/
char *GetBuffer(xCommon *Common, long Size)
{
    void *Pointer;
    
    if (Common->ToFreeTop >= ToFreeMAX)
    {
        Common->Type = 'E';
        Common->Long = 915;
        Common->ErrorType = -3;
        return NULL;
    }    
    
    Pointer = GlobalAlloc(GMEM_FIXED,Size);
    if (Pointer == NULL)
    {
        Common->Type = 'E';
        Common->Long = 913;
        Common->ErrorType = -3;
        return NULL;
    }   
    Common->ToFree[Common->ToFreeTop] = Pointer;
    Common->ToFreeTop++; 
    return Pointer;
}

	
/****************************************************************************
*                                                                           *
*    CompareStrings(stringA, stringB, NoCase)                               *
*                                                                           *
*     Compare two strings.  Trailing spaces are not considered              *
*     - strings that are equal except for trailing spaces are equal.)       *
*                                                                           *
*     If NoCase == TRUE, we do a case-insensitive sort.                     *
*                                                                           *
*     Returns  0 if the strings are equal.                                  *
*              1 if A is less than B                                        *
*              2 if A is greater than B.                                    *
*                                                                           *
*                                                                           *
****************************************************************************/
long CompareStrings(char *A, char *B, BOOL NoCase)
{
	unsigned char Ac;
	unsigned char Bc;
    
    if (NoCase)  // Case insensitive compares
    {
	    while (1) 
        {
		    Ac = (unsigned char) LOWERCASE(*A++);
		    Bc = (unsigned char) LOWERCASE(*B++);
		    if (Ac == '\0') break;
		    if (Bc == '\0') break;
		    if (Ac > Bc) return 2;
		    if (Ac < Bc) return 1;
	    }
        //  If string 1 ended... search the remainder of string 2.
	    //  Anything other than a space here means string 1 is smaller.
	    if (Ac == '\0') 
	    {
            while(1)  
            {
			    if (Bc == '\0') return 0;
			    if (Bc != ' ')  return  1;
		        Bc = (unsigned char) LOWERCASE(*B++);
            }
        }
        //  If string 2 ended, search the remainder of string 1.
        //  Anything other than a space means string 1 is larger.
        while (1)
        {
		    if (Ac == '\0') return 0;
		    if (Ac != ' ')  return 2;
		    Ac =(unsigned char) LOWERCASE(*A++);
	    }
    }
    
    
    // Case sensitive compare

	while (1) 
	{
		Ac = (unsigned char) *A++;
		Bc = (unsigned char) *B++;
		if (Ac == '\0') break;
		if (Bc == '\0') break;
		if (Ac > Bc) return 2;
		if (Ac < Bc) return 1;
	}
    //  If string 1 ended... search the remainder of string 2.
	//  Anything other than a space here means string 1 is smaller.
	if (Ac == '\0') 
	{
        while(1)  
        {
			if (Bc == '\0') return 0;
			if (Bc != ' ') return  1;
		    Bc = (unsigned char) *B++;
        }
    }
    //  If string 2 ended, search the remainder of string 1.
    // Anything other than a space means string 1 is larger.
	while (1) 
	{
		if(Ac == '\0') return 0;
		if(Ac != ' ') return 2;
		Ac =(unsigned char) *A++;
	}		
}


/****************************************************************************
*                                                                           *
*    Compare long to string.                                                *
*                                                                           *
*     converts the string to an integer, if it can, and does an integer     *
*     compare.  If the string is not a valid integer, then we return an     *
*     indication that the long is less than the string.                     *
*                                                                           *
*     Returns  0 if the values are equal.                                   *
*              1 if the long is less than the string.                       *
*              2 if the long is greater than the string.                    *
*                                                                           *
****************************************************************************/
long CompareLong2String(long L, char *S)
{

    BOOL Negative;
	long N;
    char C;
    
    N = 0;
    Negative = FALSE;
    
    while(*S == ' ') S++;
    
    switch(*S)
    {
       case '-':
       {
           Negative = TRUE;
           S++;
           break;
       }
       case '+':
       {
           S++;
           break;
       }
    }
    while (1)
    {
        C = *S++;
        if (C == '\0') break;
        if (C < '0' || C > '9') return 1;
        N = (N*10) + (C - '0');
    }
    
    if (Negative) N = -N;
    if (L < N) return 1;
    if (L > N) return 2;
    return 0;
}
       
/****************************************************************************
*                                                                           *
*    Compare string to long.                                                *
*                                                                           *
*     converts the string to an integer, if it can, and does an integer     *
*     compare.  If the string is not a valid integer, then we return an     *
*     indication that the string is greater than the long.                  *
*                                                                           *
*     Returns  0 if the values are equal.                                   *
*              1 if the string is less than the long.                       *
*              2 if the string is greater than the long.                    *
*                                                                           *
****************************************************************************/
long CompareString2Long(char *S, long L)
{

    BOOL Negative;
	long N;
    char C;
    
    N = 0;
    Negative = FALSE;
    
    while(*S == ' ') S++;
    
    switch(*S)
    {
       case '-':
       {
           Negative = TRUE;
           S++;
           break;
       }
       case '+':
       {
           S++;
           break;
       }
    }
    while (1)
    {
        C = *S++;
        if (C == '\0') break;
        if (C < '0' || C > '9') return 2;
        N = (N*10) + (C - '0');
    }
    if (Negative) N = -N;
    if (L > N) return 1;
    if (L < N) return 2;
    return 0;
}
 
/**************************************************************************
*                                                                         *
*    Convert a string to a long integer.                                  *
*                                                                         *
**************************************************************************/
long AtoL(char *A)
{
    long L;
    char C;
    
    L = 0;
    while (1)
    {
       C = *A++;
       if (C < '0') break;
       if (C > '9') break;
       L = (L * 10) + (C - '0');
    }
    return L;
}




/**************************************************************************
*                                                                         *
*    Put a string value into an array variable.                           *
*    Is passed the array variable pointer, and the string.                *
*                                                                         *
*    Returns TRUE if successful.                                          *
*                                                                         *
*    If the operation fails, puts an error indicator in the Common block  *
*    and returns FALSE.                                                   *
*                                                                         *
**************************************************************************/
DWORD PutArrayString(xCommon *Common, LPARRVAR ArrayItem, char *TheString) 
{  
    DWORD Address;
    
    Address = (DWORD) (*Common->lpfnDllServices)(Common->hBatData, 42, TheString, "", 0, 0);
    if ( Address == 0)
    {
       Common->Type = 'E';
       Common->ErrorType = -3;
       Common->Long = 104;
    } 
    else
    {
       ArrayItem->Ainittype = VARTYPE_STRING;  
       ArrayItem->Astrptr = Address;
    }
    return Address;
}

/**************************************************************************
*                                                                         *
*   Fetch the string value from an array item.                            *
*                                                                         *
**************************************************************************/
LPSTR GetArrayString(xCommon *Common, LPARRVAR ArrayItem)
{
  LONG     nResult;

  nResult = (*Common->lpfnDllServices)(Common->hBatData, 41, "", "", 0, ArrayItem->Astrptr);
  if (!(nResult))
  {
       Common->Type = 'E';
       Common->ErrorType = -3;
       Common->Long = 103;
       return NULL;
  }  
  return (LPSTR) (DWORD) nResult;
}


/***************************************************************************
*                                                                          *
*  Get the type code of an array item                                      *
*                                                                          *
***************************************************************************/
long ArrayItemType(LPARRVAR ArrayItem)
{
    long T;
    T = ArrayItem->Ainittype;
    if (T & VARTYPE_STRING) return VARTYPE_STRING;
    if (T & VARTYPE_INT) return VARTYPE_INT;
    if (T & VARTYPE_FLOATNUM) return VARTYPE_FLOATNUM;
    return -1;
}


/***************************************************************************
*                                                                          *
*   ArrayElement(Array, x1, x2, x3, x4, x5)                                *
*                                                                          *
*    Returns a pointer to the specified array element.                     *
*                                                                          *
***************************************************************************/
LPARRVAR ArrayElement(LPRESULTVAR2 arr, WORD x0, WORD x1, WORD x2, WORD x3, WORD x4)
{
  DWORD Offset, Scale, J, K;
  WORD  x[5];

  x[0] = x0;
  x[1] = x1;
  x[2] = x2;
  x[3] = x3;
  x[4] = x4;

  Offset = 0;
  for ( J = 0; J < arr->cDims; J++)
  {
    if (x[J] < 0) return NULL;                 //  subscript out of bounds
    if (x[J] >= arr->wDim[J])  return (NULL); 
    Scale = 1;
    for (K = (J + 1); K < arr->cDims; K++) Scale *= arr->wDim[K];
    Offset += (x[J] * Scale);
  }  
  return ((LPARRVAR) arr->lpBuffer + Offset);
}

/***************************************************************************
*                                                                          *
*     Debug Functions                                                      *
*                                                                          *
*      MBSN ... Display a decimal number.                                  *
*      MBSF ... Display a floating-point number                            *
*      MBSX ... Display a hexidecimal number.                              *
*      MBSS ... Display a string.                                          *
*                                                                          *
***************************************************************************/ 
 void FAR PASCAL MBSN(char *Title, long Number)
 {
     char Debug[128]; 
     wsprintf(Debug,"%d",Number);
     MessageBox(GetFocus(), Debug, Title, MB_OK | MB_SETFOREGROUND);
     return;
 }

 void FAR PASCAL MBSX(char *Title, long Number)
 {
     char Debug[128]; 
     wsprintf(Debug,"x%08lX",Number);
     MessageBox(GetFocus(), Debug, Title, MB_OK | MB_SETFOREGROUND);
     return;
 }
 
void FAR PASCAL MBSF(char *Title, double Flt)
 {
     char Debug[128];
     long I;
     long J;
     
     I = Flt;
     if (Flt < 0) Flt = 0-Flt;
     J = (Flt * 1000.0);
     wsprintf(Debug,"%d.%03.3d",I,J%1000);
     MessageBox(GetFocus(),Debug,Title,MB_OK | MB_SETFOREGROUND);
     return;
 }
 
void FAR PASCAL MBSS(char *Title, char *Text)
{
     MessageBox(GetFocus(), Text, Title, MB_OK | MB_SETFOREGROUND);
}
